""" Compatability fixes to make Python 2.7 look more like Python 3.

The general approach is to code using the common subset offered by 'six'.

The HTTPMessage class has a different interface.  This work-arounds makes the
Python 2.7 look enough like the Python 3 for the Wextracto code to work.
"""

import six

if six.PY2:

    import urllib
    import tarfile
    from httplib import HTTPMessage

    def get_content_subtype(self):
        return self.getsubtype()
    HTTPMessage.get_content_subtype = get_content_subtype

    def get_content_charset(self):
        return self.getparam('charset')
    HTTPMessage.get_content_charset = get_content_charset

    def parse_headers(fp):
        return HTTPMessage(fp, 0)

    def urlquote(name, *args, **kwargs):
        if isinstance(name, unicode):
            name = name.encode('utf-8')
        return urllib.quote(name, *args, **kwargs)

    def xzopen(cls, name, mode="r", fileobj=None, compresslevel=9, **kwargs):
        """Open lzma compressed tar archive name for reading or writing.
           Appending is not allowed.
        """
        if mode not in ("r", "w"):
            raise ValueError("mode must be 'r' or 'w'")

        try:
            from backports import lzma
            lzma.LZMAFile
        except (ImportError, AttributeError):
            raise tarfile.CompressionError("lzma module is not available")

        try:
            fileobj = lzma.LZMAFile(fileobj or name, mode)
        except (OSError, IOError):
            if mode == 'r':
                raise tarfile.ReadError("not an lzma file")
            raise

        try:
            fileobj.peek()
        except (lzma.LZMAError, EOFError):
            raise tarfile.ReadError("not an lzma file")

        try:
            t = cls.taropen(name, mode, fileobj, **kwargs)
        except IOError:
            fileobj.close()
            if mode == 'r':
                raise tarfile.ReadError("not an lzma file")
            raise
        except:
            fileobj.close()
            raise
        t._extfileobj = False
        return t

    tarfile.TarFile.xzopen = classmethod(xzopen)
    tarfile.TarFile.OPEN_METH['xz'] = 'xzopen'

else:

    import http.client
    import email.parser

    def parse_headers(fp, _class=http.client.HTTPMessage):
        """Parses only RFC2822 headers from a file pointer.

        email Parser wants to see strings rather than bytes.
        But a TextIOWrapper around self.rfile would buffer too many bytes
        from the stream, bytes which we later need to read as bytes.
        So we read the correct bytes here, as bytes, for email Parser
        to parse.

        """
        headers = []
        _MAXLINE = http.client._MAXLINE
        _MAXHEADERS = http.client._MAXHEADERS
        while True:
            line = fp.readline(_MAXLINE + 1)
            if len(line) > _MAXLINE:
                raise http.client.LineTooLong("header line")
            headers.append(line)
            if len(headers) > _MAXHEADERS:
                raise http.client.HTTPException("got more than %d headers" % _MAXHEADERS)
            if line in (b'\r\n', b'\n', b''):
                break
        # hstring = b''.join(headers).decode('iso-8859-1')
        hstring = b''.join(headers).decode('utf-8')
        return email.parser.Parser(_class=_class).parsestr(hstring)


    from six.moves.urllib_parse import quote as urlquote
    assert urlquote



""" Functions for getting responses for HTTP urls. """

from __future__ import unicode_literals, print_function
import wex.py2compat ; assert wex.py2compat
import io
import requests
from six import PY2, iteritems
from six.moves.urllib.parse import urlparse, urlunparse, parse_qsl, urlencode
from requests.sessions import merge_setting
from gzip import GzipFile
from .readable import ChainedReadable
from .http_decoder import DeflateDecoder, GzipDecoder


GZIP_MAGIC = b'\x1f\x8b'
CRLF = '\r\n'
timeout = 30.0


def remove_url_params(url, params):
    parsed = urlparse(url)
    qsl = parse_qsl(parsed.query)
    for item in iteritems(params):
        qsl.remove(item)
    replaced = parsed._replace(query=urlencode(qsl))
    return urlunparse(replaced)


def request(url, method, session=None, **kw):
    """ Makes an HTTP request following redirects. """

    import os

    user_agent = os.environ.get('WEX_USER_AGENT')

    if session is None:
        session = requests.Session()
        session.stream = True
        session.headers = {'User-Agent': user_agent} if user_agent else None

    decode_content = kw.get('decode_content', True)
    proxies = kw.get('proxies', None)
    headers = merge_setting(method.args.get('headers'), kw.get('headers'))
    context = kw.get('context', {})
    auth = merge_setting(method.args.get('auth'), kw.get('auth'))
    params = merge_setting(method.args.get('params'), kw.get('params'))

    response = session.request(
        method.name,
        url,
        allow_redirects=False,
        cookies=method.args.get('cookies', None),
        data=method.args.get('data', None),
        headers=headers,
        params=params,
        proxies=proxies,
        timeout=timeout,
        auth=auth,
    )

    if 'params' in kw:
        response.url = remove_url_params(response.url, kw['params'])

    yield readable_from_response(response, url, decode_content, context)

    redirects = session.resolve_redirects(response,
                                          response.request,
                                          proxies=proxies,
                                          stream=True,
                                          timeout=timeout)
    for redirect in redirects:
        yield readable_from_response(redirect, url, decode_content, context)


def readable_from_response(response, url, decode_content, context):
    """ Make an object that is readable by `Response`.from_file. """

    headers = io.TextIOWrapper(io.BytesIO(), encoding='utf-8', newline='\n')

    protocol = 'HTTP'
    version = '{:.1f}'.format(response.raw.version / 10.0)
    code = response.status_code
    reason = response.reason
    status_line = format_status_line(protocol, version, code, reason)
    headers.write(status_line)

    for name, value in response.headers.items():

        # Strictly speaking headers should only be iso-8859-1
        # but we've got a good chance of detecting incorrect utf-8
        # and for most headers it won't matter as they will be
        # in the common ascii subset.

        if PY2:

            try:
                name = name.decode('utf-8')
            except UnicodeDecodeError:
                name = name.decode('iso-8859-1')

            try:
                value = value.decode('utf-8')
            except UnicodeDecodeError:
                value = value.decode('iso-8859-1')

        headers.write(format_header(name.capitalize(), value))

    headers.write(format_header('X-wex-request-url', url))

    if response.url != url:
        # the URL for the response is not the same as the requested URL
        headers.write(format_header('X-wex-url', response.url))
    for name, value in context.items():
        headers.write(format_header('X-wex-context-{}'.format(name), value))
    headers.write(CRLF)
    headers.seek(0)

    # switch off urllib3 content decoding
    response.raw.decode_content = False
    content_encoding = response.headers.get('content-encoding', '').lower()
    if decode_content and content_encoding == 'gzip':
        content = GzipDecoder(response.raw)
    elif decode_content and content_encoding == 'deflate':
        content = DeflateDecoder(response.raw)
    else:
        content = response.raw

    return ChainedReadable(headers.detach(), content)


def decode(src):
    content_encoding = src.headers.get('content-encoding', '')
    declared_as_gzip = (
        src.headers.get_content_subtype() == 'x-gzip' or
        content_encoding == 'x-gzip' or
        content_encoding == 'gzip'
    )
    has_gzip_magic = (src.magic_bytes[:2] == b'\037\213')
    if declared_as_gzip and has_gzip_magic:
        src.seek(0)
        return GzipFile(fileobj=src.fp, mode='rb')
    return src


format_status_line = '{}/{} {} {}\r\n'.format
format_header = '{}: {}\r\n'.format



import threading
from .composed import wraps


class Cache(dict):

    local = threading.local()

    def __enter__(self):
        try:
            stack = self.local.stack
        except AttributeError:
            stack = self.local.stack = []
        stack.append(self)
        return self

    def __exit__(self, *exc_info):
        popped = self.local.stack.pop()
        assert popped is self

    @classmethod
    def get(cls):
        try:
            stack = cls.local.stack
        except AttributeError:
            pass
        else:
            if stack:
                return stack[-1]
        return {}


def cached(f):
    """ Decorator to cache results keyed by function arguments. """

    @wraps(f)
    def wrapper(*args):
        cache = Cache.get()
        key = (f,) + args
        try:
            return cache[key]
        except TypeError:
            return f(*args)
        except KeyError:
            result = cache[key] = f(*args)
            return result

    return wrapper
